/* Pdfviewer.ts

	Purpose:

	Description:

	History:
		Tue Oct 15 10:03:55 CST 2019, Created by rudyhuang

Copyright (C) 2019 Potix Corporation. All Rights Reserved.
 */
(() => {
const PDFJS = window['pdfjs-dist/build/pdf'];

// https://github.com/mozilla/pdf.js/blob/v2.1.266/web/pdf_link_service.js#L420
class SimpleLinkService {
	private externalLinkTarget: string | null;
	private externalLinkRel: string | null;

	public constructor() {
		this.externalLinkTarget = null;
		this.externalLinkRel = null;
	}

	public get page(): number {
		return 0;
	}

	public set page(value: number) {}

	public get rotation(): number {
		return 0;
	}

	public set rotation(value: number) {}

	public navigateTo(dest: object): void {}

	public getDestinationHash(dest: object): string {
		return '#';
	}

	public getAnchorUrl(hash: string): string {
		return '#';
	}

	public setHash(hash: string): void {}

	public executeNamedAction(action: string): void {}

	public onFileAttachmentAnnotation({ id, filename, content, }): void {}

	public cachePageRef(pageNum: number, pageRef: object): void {}
}

/**
 * A PDF file viewer.
 * <p>Only works for browsers supporting HTML5 (Firefox, Chrome, Opera, IE 11, Edge, Safari...).
 * <p>Available in ZK PE and ZK EE.
 * <p>Default {@link #getZclass}: z-pdfviewer.
 *
 * @author rudyhuang
 * @since 9.0.0
 */
zkex.pdfviewer.Pdfviewer = zk.$extends(zul.Widget, {
	_src: '',
	_activePage: 0,
	_zoom: 1,
	_zoomSelect: '1',
	_rotation: 0,
	_pageCount: 1,

	$define: {
		/**
		 * Returns the source URI of the PDF file.
		 * <p>Default: null.
		 * @return String src
		 */
		/**
		 * Sets the source URI of the PDF file.
		 * @param String src
		 */
		src(v: string) {
			if (this.desktop) {
				this._activePage = 0;
				this._loadPdf(v);
			}
		},
		/**
		 * Returns the active page (starting from 0).
		 * @return int page
		 */
		/**
		 * Sets the active page (starting from 0).
		 * @param int page
		 */
		activePage(v: number) {
			if (this.desktop) {
				let container = this.$n('container');
				container.scrollTop = 0;
				container.scrollLeft = 0;
				this._getPage(v + 1);
				this._updateToolbar();
			}
		},
		/**
		 * Returns the zoom level.
		 * <p>Default: 1.0.
		 * @return double zoom
		 */
		/**
		 * Sets the zoom level.
		 * @param double zoom
		 */
		zoom(v: number) {
			this._zoomSelect = '' + v;
			if (this.desktop) {
				this._rerenderPage();
				this._updateZoomDropdown(this._zoomSelect);
			}
		},
		/**
		 * Returns the rotation angle.
		 * <p>Default: 0.
		 * @return int rotation
		 */
		/**
		 * Sets the rotation angle.
		 * @param int rotation rotation angle. Only 0, 90, 180 and 270 are accepted.
		 */
		rotation() {
			if (this.desktop) {
				let promise = this._rerenderPage();
				if (promise) {
					promise.then(() => this._updateZoomFromSelect(this._zoomSelect));
				}
			}
		}
	},

	bind_() {
		this.$supers(zkex.pdfviewer.Pdfviewer, 'bind_', arguments);
		this.domListen_(this.$n('toolbar-zoom'), 'onChange', '_zoomTo')
			.domListen_(this.$n('toolbar-page-active'), 'onChange', '_navPageTo');
		jq(this.$n()).on('mousewheel', this.proxy(this._onWheel));
		zWatch.listen({onSize: this});
		this._updateZoomDropdown(this._zoomSelect);
		if (this._src)
			this._loadPdf(this._src);
	},
	unbind_() {
		zWatch.unlisten({onSize: this});
		jq(this.$n()).off('mousewheel', this.proxy(this._onWheel));
		this.domUnlisten_(this.$n('toolbar-page-active'), 'onChange', '_navPageTo')
			.domUnlisten_(this.$n('toolbar-zoom'), 'onChange', '_zoomTo');
		this.$supers(zkex.pdfviewer.Pdfviewer, 'unbind_', arguments);
	},

	onSize() {
		this._updateZoomFromSelect(this._zoomSelect);
	},

	/**
	 * Returns the number of pages.
	 * Note: there is at least one page even no item at all.
	 */
	getPageCount(): number {
		return this._pageCount;
	},

	onChildAdded_() {
		this.$supers('onChildAdded_', arguments);
		let toolbar = this.$n('toolbar');
		if (toolbar) toolbar.style.display = 'none';
	},
	onChildRemoved_() {
		this.$supers('onChildRemoved_', arguments);
		let toolbar = this.$n('toolbar');
		if (toolbar) toolbar.style.display = 'block';
	},
	insertChildHTML_(child: zk.Widget, before: zk.Widget | null, desktop: zk.Desktop) {
		jq(this.$n('container')).before(child.redrawHTML_());
		child.bind(desktop);
	},

	doClick_(evt: zk.Event) {
		switch (evt.domTarget) {
			case this.$n('toolbar-first'):
				this._navPageBy(-Infinity);
				evt.stop();
				break;
			case this.$n('toolbar-prev'):
				this._navPageBy(-1);
				evt.stop();
				break;
			case this.$n('toolbar-next'):
				this._navPageBy(1);
				evt.stop();
				break;
			case this.$n('toolbar-last'):
				this._navPageBy(Infinity);
				evt.stop();
				break;
			case this.$n('toolbar-zoom-in'):
				this._zoomBy(1);
				evt.stop();
				break;
			case this.$n('toolbar-zoom-out'):
				this._zoomBy(-1);
				evt.stop();
				break;
			case this.$n('toolbar-rotate-left'):
				this._rotateBy(false);
				evt.stop();
				break;
			case this.$n('toolbar-rotate-right'):
				this._rotateBy(true);
				evt.stop();
				break;
			case this.$n('toolbar-fullscreen'):
				this.toggleFullscreen();
				evt.stop();
				break;
		}
	},
	doKeyDown_(evt: zk.Event) {
		this.$supers('doKeyDown_', arguments);
		if (this._isEnabledKeyboardMouseEvent()) {
			let key = evt.domEvent.key;
			switch (key) {
				case 'ArrowUp':
				case 'ArrowLeft':
				case 'Up':
				case 'Left':
				case 'PageUp':
					this._navPageBy(-1);
					evt.stop();
					break;
				case 'ArrowDown':
				case 'ArrowRight':
				case 'Down':
				case 'Right':
				case 'PageDown':
					this._navPageBy(1);
					evt.stop();
					break;
				case 'Home':
					this._navPageBy(-Infinity);
					evt.stop();
					break;
				case 'End':
					this._navPageBy(Infinity);
					evt.stop();
					break;
			}
		}
	},
	_isEnabledKeyboardMouseEvent() {
		let zkContainer = zk(this.$n('container'));
		return !zkContainer.hasVScroll() && !zkContainer.hasHScroll();
	},
	_onWheel(evt: JQueryEventObject, delta: number) {
		if (this._isEnabledKeyboardMouseEvent()) {
			this._navPageBy(delta < 0 ? 1 : -1);
			evt.stopPropagation();
			evt.preventDefault();
		}
	},
	_navPageTo(evt: zk.Event) {
		let domTarget = evt.domTarget as HTMLInputElement;
		if (domTarget.validity.valid) {
			let newPageNum = domTarget.value;
			if (/[\d+]/.test(newPageNum)) {
				this._navPageTo0(zk.parseInt(newPageNum) - 1);
			}
		}
	},
	_navPageTo0(newActivePage: number) {
		if (!this._pdf) return;
		if (newActivePage >= 0 && newActivePage < this._pageCount) {
			if (this._activePage != newActivePage) {
				this.setActivePage(newActivePage);
				this.fire('onPaging', newActivePage);
			}
		}
	},
	_navPageBy(delta: number) {
		let newActivePage: number;
		switch (delta) {
			case -Infinity:
				newActivePage = 0;
				break;
			case Infinity:
				newActivePage = this._pageCount - 1;
				break;
			default:
				newActivePage = this._activePage + delta;
		}
		this._navPageTo0(newActivePage);
	},
	_zoomTo(evt: zk.Event) {
		if (!this._pdf) return;
		let zoomSelect = evt.domTarget as HTMLSelectElement;
		this._zoomSelect = zoomSelect.value;
		this._updateZoomFromSelect(this._zoomSelect);
	},
	_zoomBy(level: number) {
		if (!this._pdf) return;
		let oldZoom = this._zoom,
			newZoom = oldZoom + (level < 0 ? -0.1 : 0.1);
		if (newZoom > 0) {
			this.setZoom(newZoom);
			this.fire('onZoom', {zoom: this._zoom});
		}
	},
	_updateZoomFromSelect(zoom: string) {
		let zoomNumber = parseFloat(zoom);
		if (isNaN(zoomNumber)) {
			let container = this.$n('container'),
				containerHeight = container.offsetHeight,
				containerWidth = container.offsetWidth,
				page = this.$n('page'),
				pageHeight = page.offsetHeight,
				pageWidth = page.offsetWidth,
				oldZoom = this._zoom,
				scrollbarSize = jq.scrollbarWidth(),
				scrollPadding: number;
			switch (zoom) {
				case 'fit-page-width':
					scrollPadding = (pageHeight / pageWidth) > (containerHeight / containerWidth) ? scrollbarSize : 0;
					zoomNumber = (containerWidth - scrollPadding) / pageWidth * oldZoom;
					break;
				case 'fit-page-height':
					scrollPadding = (pageWidth / pageHeight) > (containerWidth / containerHeight) ? scrollbarSize : 0;
					zoomNumber = (containerHeight - scrollPadding) / pageHeight * oldZoom;
					break;
				default:
					throw 'Unknown zoom: ' + zoom;
			}
		}
		if (zoomNumber != this._zoom) {
			this._zoom = zoomNumber;
			this._rerenderPage();
			this.fire('onZoom', {zoom: this._zoom});
		}
	},
	_updateZoomDropdown(zoom: string) {
		let selectZoom = this.$n('toolbar-zoom') as HTMLSelectElement,
			options = selectZoom.options,
			zoomNumber = parseFloat(zoom),
			zoomValue = isNaN(zoomNumber) ? zoom : '' + parseFloat(zoomNumber.toPrecision(12)), // Precision issue ref: https://github.com/camsong/blog/issues/9
			found = false;
		for (let i = 0, l = options.length; i < l; i++) {
			let option = options[i];
			if (option.value == zoomValue) {
				option.selected = true;
				found = true;
				continue;
			}
			option.selected = false;
		}
		if (!found) {
			let customZoom = this.$n('toolbar-zoom-custom') as HTMLOptionElement;
			customZoom.selected = true;
			customZoom.textContent = `${Math.round(zoomNumber * 100)}%`;
		}
	},
	_rotateBy(clockwise: boolean) {
		if (!this._pdf) return;
		let oldRotation = this._rotation,
			newRotation = (oldRotation + (clockwise ? 90 : -90)) % 360;
		if (newRotation < 0) newRotation = 270;
		this.setRotation(newRotation);
		this.fire('onRotate', {rotation: newRotation});
	},

	// 1 based page number
	_getDisplayedPageNumber(): number {
		return this._activePage + 1;
	},
	_loadPdf(src: string) {
		this._destroyPdf();
		PDFJS.getDocument(src).promise.then(pdf => {
			let pageCount = pdf.numPages,
				pageCountLabel = '' + pageCount;
			this._pdf = pdf;
			this._pageCount = pageCount;
			this._getPage(this._getDisplayedPageNumber(), true);
			(this.$n('toolbar-page-active') as HTMLInputElement).max = pageCountLabel;
			this.$n('toolbar-page-total').textContent = pageCountLabel;
			this._updateToolbar();
		}, err => {
			zk.error(err.message);
			this._destroyPdf();
			this._cleanContent();
		});
	},
	_destroyPdf() {
		if (this._pdf) {
			this._pdf.destroy();
			this._pdf = null;
		}
	},
	_getPage(pageNum: number, pdfFirstRender = false): Promise<void> | void {
		if (!this._pdf) return;
		if (this._pageRendering) {
			this._pendingPageNum = pageNum;
			return;
		}

		this._pageRendering = true;
		return this._pdf.getPage(pageNum).then(page => {
			let viewport = page.getViewport({scale: this._zoom, rotation: this._rotation}),
				viewportHeight = viewport.height,
				viewportWidth = viewport.width,
				canvas = this.$n('content') as HTMLCanvasElement,
				pageLayer = this.$n('page');
			pageLayer.style.height = jq.px0(viewportHeight);
			pageLayer.style.width = jq.px0(viewportWidth);
			[canvas.height, canvas.width] = [viewportHeight, viewportWidth];
			return page.render({
					canvasContext: canvas.getContext('2d'),
					viewport: viewport
				})
				.promise
				.then(() => page.getTextContent({}))
				.then(textContent => {
					let textLayer = this.$n('text-layer'),
						child;
					while ((child = textLayer.lastChild)) {
						textLayer.removeChild(child);
					}
					return PDFJS.renderTextLayer({
						textContent: textContent,
						container: textLayer,
						viewport: viewport,
						textDivs: []
					}).promise;
				})
				.then(() => page.getAnnotations().then(annotations => {
					let annotationLayer = this.$n('annotation-layer'),
						child;
					while ((child = annotationLayer.lastChild)) {
						annotationLayer.removeChild(child);
					}
					PDFJS.AnnotationLayer.render({
						annotations: annotations,
						linkService: this.$class.SIMPLE_LINK_SERVICE,
						div: annotationLayer,
						viewport: viewport.clone({dontFlip: true}),
						page: page
					});
				}))
				.then(() => {
					this._pageRendering = false;
					if (pdfFirstRender)
						this.fire('onRender', {pageCount: this._pageCount});
					let pendingPage = this._pendingPageNum;
					if (pendingPage) {
						this._pendingPageNum = null;
						this._getPage(pendingPage);
					}
				});
		}).catch(e => {
			this._pageRendering = false;
			if (pdfFirstRender)
				this.fire('onRender', {pageCount: this._pageCount});
			zk.error(e.message);
		});
	},
	_rerenderPage(): Promise<void> | void {
		return this._getPage(this._getDisplayedPageNumber());
	},
	_cleanContent() {
		let canvas = this.$n('content') as HTMLCanvasElement,
			context = canvas.getContext('2d');
		if (context)
			context.clearRect(0, 0, canvas.width, canvas.height);
	},
	_updateToolbar() {
		let isFirstPage = this._activePage == 0,
			isLastPage = this._activePage == this._pageCount - 1;
		(this.$n('toolbar-first') as HTMLButtonElement).disabled = isFirstPage;
		(this.$n('toolbar-prev') as HTMLButtonElement).disabled = isFirstPage;
		(this.$n('toolbar-next') as HTMLButtonElement).disabled = isLastPage;
		(this.$n('toolbar-last') as HTMLButtonElement).disabled = isLastPage;
		(this.$n('toolbar-page-active') as HTMLInputElement).value = '' + this._getDisplayedPageNumber();
	},
	/**
	 * Toggles full-screen mode.
	 *
	 * Due to a specification limit, this method must be called
	 * while responding to user interaction (i.g. event handlers);
	 * otherwise it will fail.
	 */
	toggleFullscreen() {
		let n = this.$n(),
			fullscreenElement = this._fullscreenElement();
		if (n) {
			if (fullscreenElement) {
				this._exitFullscreen();
			} else {
				this._requestFullscreen(n);
			}
        }
	},
	_fullscreenElement(): Element | null {
		return document.fullscreenElement
			|| document['msFullscreenElement']
			|| document['mozFullScreenElement']
			|| document['webkitFullscreenElement'];
	},
	_requestFullscreen(elem: Element) {
		let fullscreenFn = elem.requestFullscreen
			|| elem['msRequestFullscreen']
			|| elem['mozRequestFullScreen']
			|| elem['webkitRequestFullscreen'];
		if (fullscreenFn)
			fullscreenFn.call(elem);
	},
	_exitFullscreen() {
		let exitFullscreenFn = document.exitFullscreen
			|| document['msExitFullscreen']
			|| document['mozCancelFullScreen']
			|| document['webkitExitFullscreen'];
		if (exitFullscreenFn)
			exitFullscreenFn.call(document);
	}
}, {
	WORKER_URI: zk.resourceURI('/js/zkex/pdfviewer/ext/pdf.worker.js'),
	SIMPLE_LINK_SERVICE: new SimpleLinkService(),
});

// pdj.js worker setting
PDFJS.GlobalWorkerOptions.workerSrc = zkex.pdfviewer.Pdfviewer.WORKER_URI;
})();
